package Data::Cache::Folder;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Cache::Folder - Object cache implemented above regular file storage

=head1 SYNOPSIS

 use Data::Cache::Folder;

 # Instantinate data cache in given folder (uses default TTL)
 my $cache = Data::Cache::Folder->new('./spool/cache');

 # Instantinate data cache in given folder with optional TTL (in seconds)
 my $cache = Data::Cache::Folder->new('./spool/cache', 60*60*24*7);

 # Instantinate data cache in given folder with optional TTL (in seconds).
 # Automatically perform cache cleanup
 my $cache = Data::Cache::Folder->new('./spool/cache', 60*60*24*7, 1);

 my $testa = ['array1', 'test', 'values'];
 my $testb = {a => '1', b => '1', c => '3'};

 # Store objects into cache
 my $errora = $cache->cache('key1', $testa);
 my $errorb = $cache->cache('key2', $testb);

 # Retrieve objects from cache
 my $test1 = $cache->fetch('key1');
 my $test2 = $cache->fetch('key2');

 # Clear all items from cache with TTL set up during instantination
 my $result = $cache->clear();
 # Clear all items from cache
 my $result = $cache->clear(0);
 # Clear all items from cache older than given TTL (in seconds)
 my $result = $cache->clear(60*60);

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Data::Cache(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Cwd;
use Storable;
use File::Find;

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Data::Cache;
use Value::Convertor;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant TTL_DAY  => 60*60*24;
use constant TTL_WEEK => TTL_DAY() * 7;

use constant DFLT_TTL => TTL_WEEK() * 4;

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Data::Cache');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Data.Cache.Folder instance...
}

=item clear(;$) [PUBLIC]

  Usage     : my $object = $instance->clear($timestamp)
  Purpose   : Clear data cache
  Arguments : INTEGER $ttl - time to live in seconds, clear everyting older [OPTIONAL]
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or in case of error

=cut

sub clear(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ttl,) = @_;
    $ttl = (defined($ttl))?$ttl:$self->{TTL};

    my @file_list = ();      # List of the files in cache
    my $now       = time();  # Get current time
    my $cnt       = 0;       # Number of removed files

    # Find all files in cache
    find ( { wanted => sub {
            my $file = $File::Find::name;
            if ($file =~ /(.*)/) { $file = $1; }
            if ( -f $file ) { push (@file_list, $file); }
        }, untaint => 1 }, $self->{CACHE});

    # Remove files older than TTL
    for my $file (@file_list) {
        my @stats = stat($file);
        if ($now - $stats[9] > $ttl) {
            $cnt++;
            unlink $file;
        }
    }

    return "Deleted $cnt cache files older than " . Value::Convertor->duration_to_string($ttl);
}

=item fetch($) [PUBLIC]

  Usage     : my $object = $instance->fetch($key)
  Purpose   : Fetch data object with given key from cache
  Arguments : STRING $key - object unique key [MANDATORY]
  Returns   : MIXED $object on SUCCESS, undef on FAILURE
  Throws    : Dies, if not invoked on instance or given invalid arguments

=cut

sub fetch($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($key,) = @_;
    confess "Cache key must be given as first argument" unless $key;
    my $name = $self->_generate_file_name($key);

    return undef unless -f $name;

    Log::Loger->notice(__PACKAGE__, "Fetching '$key' from folder cache");
    if ($self->{CTYPE} eq 'bin') {
        return Storable::retrieve($name);
    }
    else {
        open (my $cf, $name) or die "Can`t open cache file '$name'\n";
        my $json = join("", <$cf>);
        close($cf);
        return decode_json($json);
    }
}

=item fetch_newer($$) [PUBLIC]

  Usage     : my $object = $instance->fetch_newer($key, $timestamp)
  Purpose   : Fetch data object with given key from cache, but the object must be
              newer or equal given timestamp
  Arguments : STRING $key - object unique key [MANDATORY]
              INTEGER $timestamp - timestamp [MANDATORY]
  Returns   : MIXED $object on SUCCESS, undef on FAILURE
  Throws    : Dies, if not invoked on instance or given invalid arguments

=cut

sub fetch_newer($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($key, $timestamp) = @_;
    confess "Cache key must be given as first argument" unless $key;
    confess "Cache object timestamp must be given as second argument" unless $timestamp;

    # Generate file name and check for it`s existence
    my $name = $self->_generate_file_name($key);
    return undef unless -f $name;

    # Return the object from cache, only in case it is newer than given timestamp
    my $ts = (stat($name))[9];
    if ($self->{CTYPE} eq 'bin') {
        return Storable::retrieve($name) if $ts >= $timestamp;
    }
    else {
        open (my $cf, $name) or die "Can`t open cache file '$name'\n";
        my $json = join("", <$cf>);
        close($cf);
        return decode_json($json);
    }

    # Remove the obsolete object from cache
    unlink $name;
    return undef;
}

=item cache($$;$) [PUBLIC]

  Usage     : my $object = $instance->cache($key, $object, $overwrite)
  Purpose   : Store data object under given key to cache
  Arguments : STRING $key - object unique key [MANDATORY]
              MIXED $object - object to be stored
  Returns   : STRING $error on FAILURE, EMPTY STRING on SUCCESS
  Throws    : Dies, if not invoked on instance or given invalid arguments

=cut

sub cache($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($key, $object, $overwrite) = @_;
    confess "Cache key must be given as first argument" unless $key;
    confess "Stored object must be given as argument" unless $object;
    confess "Stored object must be given via reference" unless ref $object;
    my $name = $self->_generate_file_name($key);

    # Handle the instances the cache object already exists
    if (-f $name) {
        if ($overwrite) {
            unlink $name;
        }
        else {
            return "Cache item for key '$key' already exists";
        }
    }

    if ($self->{CTYPE} eq 'bin') {
        # Store the object to the binary cache.
        Storable::store($object, $name);
    }
    else {
        # Store the object to the JSON cache.
        my $json = JSON->new->utf8->allow_blessed->convert_blessed->encode($object);
        open(my $hnd, ">$name") or die "Cannot open cache file '$name' for writing\n";
        print $hnd $json;
        close($hnd);
    }

    # Make sure the cache file access permissions are normalized
    my $mode = 0664; chmod($mode, $name)
        or die "Unable to change permissions of the cache file '$name' to '$mode'\n";

    return '';
}


#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Data::Cache::Folder instance
#  Arguments : STRING $cache_folder - name of the cache folder [MANDATORY]
#              INTEGER $ttl - TTL of cache items [OPTIONAL]
#              BOOL $clear - clear old items from cache upon startup [OPTIONAL]
#  Returns   : Data::Cache::Folder $self reference
#  Throws    : Dies, if not invoked on instance or given invalid arguments
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Data.Cache.Folder instance...
    my ($cache_folder, $ttl, $clear, $ctype) = @_;
    confess "Cache folder is expected to be given as first argument" unless $cache_folder;
    confess "Cache folder must be writable directory" unless -d $cache_folder and -w $cache_folder;
    confess "Cache TTL must be integer and greater than zero" if $ttl and not int($ttl);

    $ctype = 'bin' unless $ctype;

    # Cleanup cache folder name
    $cache_folder = Cwd::abs_path($cache_folder);
    $cache_folder =~ s\/$\\;
    if ($cache_folder =~ /(.*)/) { $cache_folder = $1; }

    # Setup internal configurations
    $self->{CACHE} = $cache_folder;
    $self->{TTL}   = ($ttl)?int($ttl):DFLT_TTL();
    $self->{CTYPE} = ($ctype eq 'json')?'json':'bin';

    # Clear old values from cache
    $self->clear() if $clear;

    return $self;
}

# _generate_file_name() [PROTECTED]
#
#  Usage     : my $name = $instance->_generate_file_name($key);
#  Purpose   : Generate cache file name for given cache key
#  Arguments : STRING $key - cache key [MANDATORY]
#  Returns   : STRING $name of the cache file
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _generate_file_name($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($key,) = @_;

    my $name = $self->{CACHE} . "/$key." . $self->{CTYPE};
    if ($name =~ /(.*)/) { $name = $1; }
    return $name;
}

=pod

=back

=cut

1;
