package Data::Convertor;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Convertor - Short module description

=head1 SYNOPSIS

 use Data::Convertor;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use JSON;

#-- Custom application modules ------------------------------------------------#
use Value::Convertor;
use Mentat::Message::Factory;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Data::Convertor->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Data::Convertor reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my %config = @_;

    my $self = bless ({}, $class);

    $self->{MAP} = $self->_load_json_file($config{map_file}) if $config{map_file};

    return $self->_init(@_);
}

=item convert($;$$) [PUBLIC]

  Usage     : my $result = $instance->convert($message)
  Purpose   : Convert whole Mentat::Message to different Mentat message class or data structure
  Arguments : Mentat::Message $message - message to be converted
  Returns   : REFERENCE containing the converted data structure
  Throws    : Croaks, if not invoked on object

=cut

sub convert($;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message, $class, $iterspec) = @_;
    confess "Message must be given as first argument" unless defined($message);

    my @tmp          = split(/:+/, blessed($self));
    my $target_class = lc(pop @tmp);

    if (blessed($message)) {
        $class = lc($message->class()) unless $class;
    }
    else {
        $class   = 'tree' unless $class;
        $message = Data::Tree->new($message)
    }

    # Perhaps the message is already in correct class
    return $message if $class eq $target_class;

    # Determine the name of the actual conversion helper method
    my $conversion_helper = lc("convert_node_$class");
    confess "Cannot convert message class '$class'" unless $self->can($conversion_helper);

    # Prepare storage for message rules
    my $rules = [];

    # Iterate through whole message
    my ($src_mpath, $value) = $message->iterate(undef, undef, undef, $iterspec);
    while ($src_mpath) {
        # CLEANUP - Get rid of the empty nodes
        unless (defined $value) {
            print STDERR "Skipping empty node on path '$src_mpath'\n" if $DEVEL;
        }
        # CLEANUP - Get rid of the text nodes containing whitespace formating
        elsif ($value =~ /^\s*$/) {
            print STDERR "Skipping empty value on path '$src_mpath'\n" if $DEVEL;
        }
        else {
            print STDERR "CONVERTING NODE: '$src_mpath':'$value'\n" if $DEVEL;

            # Perform the conversion of current node
            my @r = $self->$conversion_helper($message, $src_mpath, $value);
            # Push any obtained message rules for the target message
            push(@$rules, @r) if @r;
        }

        # Continue iteration
        ($src_mpath, $value) = $message->iterate();
    }

    # Finalize the conversion rules
    $self->_finalize($rules, $message);

    # Build and return the message using obtained rules
    return Mentat::Message::Factory->from_rules($rules, $target_class);
}

=item convert_value($) [PUBLIC]

  Usage     : my $result = $instance->convert_value($mpath, $value)
  Purpose   : Convert given value, which was located on given MPath, to optimal representation
  Arguments : STRING $mpath - path to value
              STRING $value - value as string
  Returns   : Value in optimal representation (INT, STRING, ...)
  Throws    : Croaks, if not invoked on object

=cut

sub convert_value($$)
{
    my $self = shift;
    confess "Method needs implementation";
}

=item convert_for_comparison($) [PUBLIC]

  Usage     : my $result = $instance->convert_for_comparison($mentat_message)
  Purpose   : Convert whole data structure to different data structure
  Arguments : Data::Tree $data - data structure to be converted
  Returns   : REFERENCE containing the converted data structure
  Throws    : Croaks, if not invoked on object

=cut

sub convert_for_comparison($)
{
    my $self = shift;
    confess "Method needs implementation";
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Data::Convertor instance
#  Arguments :
#  Returns   : Data::Convertor reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self;
}

# _load_json_file($) [PROTECTED]
#
#  Usage     : my $json = $self->_load_json_file($file);
#  Purpose   : Load JSON structure from external file
#  Arguments : STRING $file - name of the JSON file
#  Returns   : REFERENCE $json

sub _load_json_file($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($file,) = @_;
    confess "JSON file must be given as argument" unless $file;
    confess "JSON file must be ordinary file" unless -f $file;

    # Load JSON string from external file
    open (my $fh, $file) or die "Can`t open JSON file '$file'\n";
    my $json;
    while (my $line = <$fh>) {
        # Custom feature - remove comments
        next if $line =~ /^\s*(#|;|\/\/)/;
        $json .= $line;
    }
    close($fh);

    # Untaint the loaded string
    if ($json =~ /^(.+)$/s) { $json = $1; }

    # Parse out the JSON object and return it
    return decode_json($json);
}

# _map($$) [PROTECTED]
#
#  Usage     : my $json = $self->_map($section, $key);
#  Purpose   :
#  Arguments :
#  Returns   :

sub _map($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($section, $key) = @_;

    return undef unless $self->{MAP}->{$section} and $self->{MAP}->{$section}->{$key};
    return $self->{MAP}->{$section}->{$key};
}

#-------------------------------------------------------------------------------

# _finalize($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _finalize($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($rules, $message) = @_;
}


# _convert_node($$$$$;@) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _convert_node($$$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($src_mpath, $value, $conversion, $variables, $dst_mpath, @args) = @_;

    $dst_mpath = $dst_mpath || $src_mpath;
    $dst_mpath = $self->_create_dst_mpath($dst_mpath, $variables);
    $value = Value::Convertor->$conversion($value, @args);
    return ([$dst_mpath, $value]);
}

# _convert_with_conversion($$;@) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _convert_with_conversion($$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($conversion, $value, @args) = @_;

    if ($self->can($conversion)) { return $self->$conversion($value, @args); }
    else                         { return Value::Convertor->$conversion($value, @args); }
}

# _create_dst_mpath($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _create_dst_mpath($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($dst_mpath, $variables) = @_;

    print STDERR "BEFORE: $dst_mpath\n" if $DEVEL;

    $dst_mpath =~ s{
             \$                     # find a literal dollar sign
            ([a-zA-Z0-9_]+)         # find a "word" and store it in $1
    }{
        if (defined $variables->{$1}) {
            $variables->{$1};       # expand global variables only
        }
        else {
            "";
        }
    }egx;
    $dst_mpath =~ s/\[\]/[1]/g;

    print STDERR "AFTER: $dst_mpath\n" if $DEVEL;

    return $dst_mpath;
}

=pod

=back

=cut

1;
