package Data::Filter::Compiler;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Filter::Compiler - Compile/compress filter tree objects into optimal data representation

=head1 SYNOPSIS

 use Data::Filter::Compiler;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Data::Filter::Traverser(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #--- DEVELOPMENT ONLY ---#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Data::Filter::Traverser;
use Data::Core qw(:c_dtypes :c_opers);
use Data::Filter::Factory;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Data::Filter::Traverser');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item compile($$) [PUBLIC]

  Usage     : my $result = $instance->compile($filter);
  Purpose   : Compile/compress filter tree objects into optimal data representation
  Arguments : Data::Filter $filter - query filter tree
  Returns   : Data::Filter compiler filter
  Throws    : Croaks, if not invoked on object

=cut

sub compile($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;

    return $filter->traverse($self);
}

#-------------------------------------------------------------------------------

=item FILTER CALLBACKS [PUBLIC]

  Usage     : return $traverser->$name(@args);
  Purpose   : Rule tree traversal callbacks, for each rule class one callback method
  Arguments : Depend on particular callback
  Returns   : Any result from callback will be passed to the parent node callback as argument
  Throws    : Croaks, if invoked on object
  Comments  : These methods are called from Data::Filter::traverse method
              during filter tree traversal. Arguments are dependent on

=cut

sub constant($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;
    return $filter;
}
sub list($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;
    return $filter;
}
sub existence($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;
    return $filter;
}
sub variable($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter,) = @_;
    return $filter;
}

# ---
sub operationUnary($$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, $subfilter, $result) = @_;

    # If operation operand is constant, evaluate the operation
    if ($result->isa('Data::Filter::Rule::Constant')) {
        my $evaluation = $self->evaluate_n_unr($operation, $result);

        return Data::Filter::Factory->rule('const', $evaluation) unless (ref $evaluation eq 'ARRAY');
        return Data::Filter::Factory->rule('list', $evaluation);
    }
    return $filter;
}
sub operationBinary($$$$$$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter, $operation, $l_subfilter, $r_subfilter, $l_result, $r_result) = @_;

    # Change the filter`s operands to the processed ones
    $filter->left_operand($l_result);
    $filter->right_operand($r_result);

    # If both operation operands are constants, evaluate the operation
    if (($l_result->isa('Data::Filter::Rule::Constant') or $l_result->isa('Data::Filter::Rule::List')) and
        ($r_result->isa('Data::Filter::Rule::Constant') or $r_result->isa('Data::Filter::Rule::List'))) {
        my $evaluator = $self->pick_evaluator($operation, $l_result, $r_result);
        my $evaluation = $self->evaluate_n_bin($evaluator, $l_result->value_raw(), $r_result->value_raw());

        return Data::Filter::Factory->rule('const', $evaluation) unless (ref $evaluation eq 'ARRAY');
        return Data::Filter::Factory->rule('list', $evaluation);
    }
    # If one of the operation operands is constant and other is variable,
    # compile/convert the constant value into optimal representation depending
    # on the datatype of the sibling operand
    elsif ($l_result->isa('Data::Filter::Rule::Constant') or $l_result->isa('Data::Filter::Rule::List')) {
        my ($rd, $rv) = ($self->resolve_datatype($r_result), $r_result->value);
        $rd = DT_REGEXP if ($operation eq OP_LIKE());
        $l_result->compile($rd);
    }
    elsif ($r_result->isa('Data::Filter::Rule::Constant') or $r_result->isa('Data::Filter::Rule::List')) {
        my ($ld, $lv) = ($self->resolve_datatype($l_result), $l_result->value);
        $ld = DT_REGEXP if ($operation eq OP_LIKE());
        $r_result->compile($ld);
    }
    return $filter;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Data::Filter::Compiler instance
#  Arguments :
#  Returns   : Data::Filter::Compiler reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self;
}

=pod

=back

=cut

1;
