package Hawat::Module;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Hawat::Module - Base class for all Hawat modules

=head1 SYNOPSIS

 use Hawat::Module;

=head1 DESCRIPTION

Base class for all Hawat administration modules. All classes implementing this
interface may be used as dynamic plugins for the hawat-cli command.

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Term::ANSIColor;

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

my %MODULES = ();

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION    = '0.1';
    $DEVEL      = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item register_module($) [PUBLIC,STATIC]

  Usage     : my $name = $module->register_module($module);
  Purpose   : Return unique name of the module
  Returns   :
  Arguments :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub register_module($)
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($module,) = @_;
    confess "Module must be given as argument" unless $module;
    confess "Module must be instance of 'Hawat::Module' class" unless blessed($module) and $module->isa('Hawat::Module');

    my $module_name = $module->get_name();
    confess "Module '$module_name' is already registered" if $MODULES{$module_name};

    return $MODULES{$module_name} = $module;
}

=item get_names() [PUBLIC]

  Usage     :
  Purpose   :
  Returns   :
  Arguments :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_names()
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    return keys(%MODULES);
}

=item get_module($) [PUBLIC]

  Usage     : my $name = $module->get_module($);
  Purpose   :
  Returns   :
  Arguments :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_module($)
{
    my $invocant = shift;
    my ($module_name,) = @_;
    confess "Module name must be given as argument" unless $module_name;

    return undef unless exists $MODULES{$module_name};
    return $MODULES{$module_name};
}

#-----------------------------------------------------------------------

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Hawat::Module->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Hawat::Module reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($configs, $writer) = @_;

    my $self = bless ({}, $class);

    $self->{CFG} = $configs;
    $self->{STD} = $writer;
    $self->{RES} = {};

    return $self->_init(@_);
}

=item out($;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub out($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message, $color) = @_;

    my $output = $self->{STD};
    print $output color $color if $color;
    print $output "$message";
    print $output color 'reset' if $color;
}

=item res() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub res()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return $self->{RES};
}


=item cfg() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub cfg()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return $self->{CFG};
}

#-------------------------------------------------------------------------------

=item get_commands() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_commands()
{
    my $self = shift;
    confess "Method needs implementation";
}

=item get_name() [PUBLIC]

  Usage     : my $name = $module->get_name();
  Purpose   : Return unique name of the module
  Returns   :
  Arguments :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_name()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Determine the name of the rule (last part of the class name)
    unless ($self->{NAME}) {
        my $name = blessed($self);
        my @tmp = split(/:+/, $name);
        $self->{NAME} = lc($tmp[$#tmp]);
    }
    return $self->{NAME};
}

=item get_help(;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub get_help(;$@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message, @command_chain) = @_;

    my %commands = $self->get_commands();

    my $help_msg = '';

    $help_msg .= "$message\n" if $message;
    $help_msg .= "Invalid command @command_chain\n" if @command_chain;

    $help_msg .= ' \'' . $self->get_name() . "' module commands:\n";
    foreach my $command_name (keys(%commands)) {
        $help_msg .= $self->_get_command_help($command_name, $commands{$command_name}, 0);
    }

    return $help_msg;
}

#-------------------------------------------------------------------------------

=item handle_command(;@) [PUBLIC]

  Usage     :
  Purpose   : Handle given command
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on class
  Comments  :
  See Also  :

=cut

sub handle_command(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my %commands = $self->get_commands();

    $self->_find_and_invoke_command(\%commands, [$self->get_name()], @_);
    $self->out("\n[DONE]\n", 'bold yellow');

    #return $self->res;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Hawat::Module instance
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Method needs implementation";
}

# _find_and_invoke_command($$@) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _find_and_invoke_command($$@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($commands, $chain, @args) = @_;

    # If there are any subcommands for this command, we have to dig deeper
    if (ref $commands eq 'HASH') {
        # Extract next command from arguments
        my $command = shift(@args);

        unless ($command) {
            if (exists $commands->{''}) {
                unshift (@args, $command) if $command;
                return $self->_invoke_command($commands->{''}, $chain, @args);
            }
            else {
                die $self->get_help("Invalid usage! You have to select one of the valid module commands.", @$chain);
            }
        }
        elsif (exists ($commands->{$command}) and $command ne '_DESC_')
        {
            return $self->_find_and_invoke_command($commands->{$command}, [@$chain, $command], @args);
        }
        elsif (exists $commands->{''})
        {
            unshift (@args, $command) if $command;
            return $self->_invoke_command($commands->{''}, [@$chain, $command], @args);
        }
        else {
            die $self->get_help("Invalid usage! Unknown command '$command'.", @$chain);
        }
    }
    # Otherwise try to execute the appropriate command with current arguments
    else {
        return $self->_invoke_command($commands, $chain, @args);
    }
}

# _invoke_command($@) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _invoke_command($@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($method_metadata, $chain, @args) = @_;

    my $method_name = $method_metadata->[0];

    die "Module '" . $self->get_name() . "' does not have command method '$method_name' implemented !!!\n" unless $self->can($method_name);
    #print "INVOKING METHOD '$method_name'" . ((@args) ? " with args '@args'" : '') . "\n";

    return $self->$method_name($chain, @args);
}

# _get_command_help($$) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _get_command_help($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($command_name, $command_metadata, $depth) = @_;

    $depth += 1;
    my $help_msg = '';
    my $command_display_name = $command_name || '[EMPTY]';
    my $indent = ('    ' x $depth);
    my $dot = ($depth == 1) ? '* ' : '- ';

    if (ref $command_metadata eq 'HASH')
    {
        $help_msg .= "$indent$dot$command_display_name - " . $command_metadata->{'_DESC_'} . "\n";
        foreach my $sub_command (keys(%{$command_metadata}))
        {
            unless ($sub_command eq '_DESC_')
            {
                $help_msg .= $self->_get_command_help($sub_command, $command_metadata->{$sub_command}, $depth);
                $help_msg .= "\n";
            }
        }
    }
    else {
        my @cmd_metadata = @$command_metadata;
        $help_msg .= "$indent$dot$command_display_name - " . $cmd_metadata[1] . "\n";
        my @args = @cmd_metadata[2..$#cmd_metadata];
        if (@args)
        {
            #$help_msg .= "$indent       Arguments:\n";
            foreach my $arg (@args)
            {
                $help_msg .= "$indent       $arg\n";
            }
        }
    }

    return $help_msg;
}

=pod

=back

=cut

1;
