package Mentat::Processor::Scheduler::Default;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Scheduler::Default - Dafault implementation of processor event scheduler

=head1 SYNOPSIS

 use Mentat::Processor::Scheduler::Default;

=head1 DESCRIPTION

This class is the default implementation of Mentat::Processor::Scheduler event
scheduler interface for Mentat::Processor main application controller.

=head1 IMPLEMENTATION DETAILS

Scheduler is using following principle for storing events is queues: references to
array references are stored in queue! This principle can is ilustrated on following
code:

 my $array = ['test', 'test'];

 my $queue = [\$array];
 my $map   = {id => \$array};

 print STDERR Dumper($queue, $map);

 # Replace the item with empty item (NOTE THE DEREFERENCE!)
 # This is used during event cancelation
 ${$map->{id}} = [];
 delete $map->{id};

 print STDERR Dumper($queue, $map);

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Mentat::Processor::Scheduler(3), Mentat::Processor(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#
use Tie::Array::Sorted;
use Tie::Array::Sorted::Lazy;

#-- Custom application modules ------------------------------------------------#
use Tie::Array::Sorted::Stable;
use Tie::Array::Sorted::Stable::Lazy;
use Mentat::Processor::Scheduler;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant TOI => 2;  # Index of timeout field within the event queue record

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Mentat::Processor::Scheduler');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item schedule($;@) [PUBLIC]

  Usage     : my $id = $instance->schedule($event, @args);
  Purpose   : Schedule new event at the end of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Croaks, if not invoked on class

=cut

sub schedule($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, @args) = @_;
    confess "Event name must be given as first argument" unless $event_name;

    my $id = $self->_id();
    my $event = [$id, $event_name, 0, @args];

    # Enqueue the reference to the event, not event itself, otherwise event
    # cancelation would not work
    push(@{$self->{EVENTS}}, \$event);
    # Record the event to the event map (for backreference during event cancelation)
    # and again use reference to the event
    $self->{EVENTS_MAP}->{$id} = \$event;

    $self->{STATISTICS}->{scheduled}++;
    return $id;
}

=item schedule_first($;@) [PUBLIC]

  Usage     : my $id = $instance->schedule_first($event, @args);
  Purpose   : Schedule new event at the beginning of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Croaks, if not invoked on class

=cut

sub schedule_first($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, @args) = @_;
    confess "Event name must be given as first argument" unless $event_name;

    my $id = $self->_id();
    my $event = [$id, $event_name, 0, @args];

    # Enqueue the reference to the event, not event itself, otherwise event
    # cancelation would not work
    unshift(@{$self->{EVENTS}}, \$event);
    # Record the event to the event map (for backreference during event cancelation)
    # and again use reference to the event
    $self->{EVENTS_MAP}->{$id} = \$event;

    $self->{STATISTICS}->{scheduled_first}++;
    return $id;
}

=item schedule_at($$;@) [PUBLIC]

  Usage     : my $id = $instance->schedule_at($event, @args);
  Purpose   : Schedule new event at specific time
  Arguments : STRING $event - event name [MANDATORY]
              INTEGER $timestamp - unix timestamp [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Croaks, if not invoked on class

=cut

sub schedule_at($$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, $timestamp, @args) = @_;
    confess "Event name must be given as first argument" unless $event_name;
    confess "Event timestamp must be given as second argument" unless $timestamp;

    my $id = $self->_id();
    my $event = [$id, $event_name, $timestamp, @args];

    # Enqueue the reference to the event, not event itself, otherwise event
    # cancelation would not work
    push(@{$self->{EVENTS_T}}, \$event);
    # Record the event to the event map (for backreference during event cancelation)
    # and again use reference to the event
    $self->{EVENTS_MAP}->{$id} = \$event;

    $self->{STATISTICS}->{scheduled_at}++;
    return $id;
}

=item next_event() [PUBLIC]

  Usage     : my $event = $instance->next_event();
  Purpose   : Fetch next event from the event queue
  Arguments : None
  Returns   : ARRAY REFERENCE $event on SUCCESS, UNDEF on FAILURE
  Throws    : Dies, if not invoked on instance

=cut

sub next_event()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $time = time();
    my ($next_time, $result);

    ### First check the time-specific scheduled events
    while ($self->{EVENTS_T}->[0]) {
        ### Check, that it is not an empty event
        if (${$self->{EVENTS_T}->[0]}->[0]) {
            ### Check, that it is time to execute the first event
            $next_time = ${$self->{EVENTS_T}->[0]}->[TOI()];
            if ($time >= $next_time) {
                $result = ${shift(@{$self->{EVENTS_T}})};
                delete $self->{EVENTS_MAP}->{$result->[0]};
                $self->{STATISTICS}->{returned_at}++;
                return $result;
            }
            ### Go check time-nonspecific scheduled events
            last;
        }
        ### It was empty event, skip and try next one
        $self->{STATISTICS}->{skipped_at}++;
        shift(@{$self->{EVENTS_T}});
        next;
    }

    ### Finally check time-nonspecific scheduled events
    while ($self->{EVENTS}->[0]) {
        ### Check first event in queue
        $result = ${shift(@{$self->{EVENTS}})};
        if($result->[0]) {
            ### It is a valid event, return it
            delete $self->{EVENTS_MAP}->{$result->[0]};
            $self->{STATISTICS}->{returned}++;
            return $result;
        }
        ### It was empty event, skip and try next one
        $self->{STATISTICS}->{skipped}++;
        next;
    }

    ### No event to process at this time
    return (undef, $next_time);
}

=item cancel_event($) [PUBLIC]

  Usage     : my $error = $instance->cancel_event($id);
  Purpose   : Cancel event with given identifier
  Arguments : MIXED $id - event identifier
  Returns   : STRING $error on FAILURE, EMPTY STRING on SUCCESS
  Throws    : Dies, if not invoked on instance

=cut

sub cancel_event($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_id,) = @_;
    confess "Event identifier must be given as first argument" unless $event_id;

    # Check if the event with given identifier exists
    return "Event with identifier '$event_id' does not exist" unless exists $self->{EVENTS_MAP}->{$event_id};

    # Replace event with empty event (we don`t know, in which queue and on what
    # position is it, so we are using references to references). We use empty
    # event because it is valid event, so it does not mess up the auto sorting
    # algorithm in Tie::Array::Sorted.
    ${$self->{EVENTS_MAP}->{$event_id}} = [0,0,${$self->{EVENTS_MAP}->{$event_id}}->[TOI()]];
    # Delete the event from the event map
    delete $self->{EVENTS_MAP}->{$event_id};

    $self->{STATISTICS}->{canceled}++;
    return '';

    # Removing elements from the middle of the array
    # splice @array, $ELEMENT_NUMBER_WE_WANT_DELETED, 1;
}

=item count(;$) [PUBLIC]

  Usage     : my $count = $instance->count();
  Purpose   : Get number of events in event queue
  Arguments : BOOL $real - don't count empty events (slower) [OPTIONAL]
  Returns   : INTEGER $count
  Throws    : Croaks, if not invoked on class
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub count(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($real,) = @_;

    # By default return count of all events in all queues (including empty events)
    return (scalar @{$self->{EVENTS}}) + (scalar @{$self->{EVENTS_T}}) unless $real;
    # Optionally don't count empty events
    return (scalar keys(%{$self->{EVENTS_MAP}}));
}

=item statistics() [PUBLIC]

  Usage     : my $statistics = $instance->statistics();
  Purpose   : Get processing statistics
  Arguments : None
  Returns   : HASH REFERENCE $statistics with various keys
  Throws    : Dies, if not invoked on instance
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub statistics()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->{STATISTICS};
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init(%) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor::Scheduler::Default instance
#  Arguments : HASH %config
#  Returns   : Mentat::Processor::Scheduler::Default reference
#  Throws    : Dies, if not invoked on instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init(%)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Initializing Mentat.Processor.Scheduler.Default instance...
    my %config = @_;

    my $type = $config{type};
    if ($type) {
        if    ($type eq 'unstable') { $type = "Tie::Array::Sorted"; }
        elsif ($type eq 'stable')   { $type = "Tie::Array::Sorted::Stable"; }
        elsif ($type eq 'lazy')     { $type = "Tie::Array::Sorted::Lazy"; }
        elsif ($type eq 'stlazy')   { $type = "Tie::Array::Sorted::Stable::Lazy"; }
    }
    else {
        $type = "Tie::Array::Sorted::Stable";
    }

    my @a;
    tie @a, $type, sub { ${$_[0]}->[TOI()] <=> ${$_[1]}->[TOI()] };

    # Queue for events not scheduled at specific time
    $self->{EVENTS} = [];
    # Queue for events scheduled at specific time
    $self->{EVENTS_T} = \@a;
    # Map EVENT_ID => EVENT (allows easy event cancellation)
    $self->{EVENTS_MAP} = {};
    # Processing statistics (for debugging, development and optimization purposes)
    $self->{STATISTICS} = {};

    return $self;
}

=pod

=back

=cut

1;
